use std::{io::Cursor, sync::mpsc, time::Instant};

use rodio::{OutputStream, OutputStreamHandle, Sink};

pub enum Cmd {
    GetTime(oneshot::Sender<f32>),
    GetIsPaused(oneshot::Sender<bool>),
    SetInput(f32),
    TogglePaused,
    Reset,
}

pub struct State {
    stream: OutputStream,
    stream_handle: OutputStreamHandle,
    last_start: Instant,
    input_time: f32,
    ellapsed_at_last_pause: f32,
    is_paused: bool,
    reached_target: bool,
    player: Option<Sink>,
}

impl State {
    pub fn initial() -> Self {
        let (stream, stream_handle) = OutputStream::try_default().unwrap();

        Self {
            stream,
            stream_handle,
            last_start: Instant::now(),
            input_time: 0.,
            ellapsed_at_last_pause: 0.,
            is_paused: true,
            reached_target: false,
            player: None,
        }
    }

    pub fn spin(mut self, commands: mpsc::Receiver<Cmd>) {
        while let Ok(cmd) = commands.recv() {
            match cmd {
                Cmd::GetTime(returns) => {
                    let mut total_ellapsed = self.ellapsed_at_last_pause;
                    if !self.is_paused {
                        total_ellapsed += self.ellapsed_since_last_start();
                    }
                    let res = self.input_time - total_ellapsed;
                    if res < 0. && !self.reached_target {
                        self.reached_target = true;
                        self.start_playing();
                    }
                    returns.send(res.max(0.)).ok(); // note : sends zero instead of negatives.
                }
                Cmd::GetIsPaused(returns) => {
                    returns.send(self.is_paused).ok();
                }
                Cmd::SetInput(inputted) => {
                    self.input_time = inputted;
                }
                Cmd::TogglePaused => {
                    self.is_paused = !self.is_paused;
                    if self.is_paused {
                        self.ellapsed_at_last_pause += self.ellapsed_since_last_start();
                    } else {
                        self.last_start = Instant::now();
                    }
                }
                Cmd::Reset => {
                    self.is_paused = true;
                    self.ellapsed_at_last_pause = 0.;
                    self.reached_target = false;
                    self.stop_playing();
                }
            }
        }
    }

    pub fn ellapsed_since_last_start(&self) -> f32 {
        Instant::now().duration_since(self.last_start).as_secs_f32()
    }

    pub fn start_playing(&mut self) {
        println!("start");
        let cursed = Cursor::new(DING);
        //        let source = Decoder::new_mp3(cursed).unwrap();
        let control = self.stream_handle.play_once(cursed).unwrap();
        control.play();
        self.player.replace(control);
    }

    pub fn stop_playing(&self) {
        println!("stop");
        if let Some(player) = &self.player {
            player.stop();
        }
    }
}

const DING: &[u8] = include_bytes!("../assets/ding.ogg");
